from copili import Pipeline
import schedule
import docker
import os
import sys
import logging

logging.basicConfig(handlers=[logging.StreamHandler()], level=logging.INFO)
log = logging.getLogger(__name__)


if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )

    sys.path.append(os.path.normpath(SCRIPT_DIR))


d = docker.DockerClient(base_url="unix://var/run/docker.sock")

yaml = """
ExmaplePipeline:
    - name: helloworld
      image_repo: hello-world
      sidecars:
        - name: redis01
          image_repo: redis
      dependencies:
        - random-logs
    - name: random-logs
      image_repo: chentex/random-logger
      command:
        - 1000
        - 4000
        - 100
      labels: 
        mylabel: True
"""
'''
yaml = """
ExmaplePipeline:
    - name: test01
      image_repo: registry-gl.connect.dzd-ev.de:443/dzdtools/pubmedsucker
      env_vars:
        CONFIGS_REDIS: {"host":"redis01"}
        CONFIGS_NEO4J: {"host":"192.168.178.69"}
      sidecars:
        - name: redis01
          image_repo: redis
"""
'''

'''
yaml = """
CovidGraphPipeline:
    - name: CREATE_FULL_TEXT_INDEXES
      image_repo: covidgraph/graph-processing-fulltext-indexes
    - name: LENS_PATENT_DATA
      image_repo: covidgraph/data-lens-org-covid19-patents
    - name: CORD19
      image_repo: covidgraph/data-cord19
      env_vars:
        CONFIGS_PAPER_BATCH_SIZE: 300
        CONFIGS_NO_OF_PROCESSES: 25
    - name: TEXT_FRAGGER
      image_repo: covidgraph/graph-processing_fragmentize_text
      dependencies:
        - CORD19
        - LENS_PATENT_DATA
    - name: TEXT_GENE_MATCH
      image_repo: covidgraph/graph-processing_text_gene_match
      dependencies:
        - CORD19
        - LENS_PATENT_DATA
        - BIOBASE
        - TEXT_FRAGGER
    - name: CLINICAL_TRIALS_GOV
      image_repo: covidgraph/data-clinical_trials_gov
    - name: HELOMICS_HETIONET
      image_repo: helomics/data_hetionet
    - name: BIOBERT
      image_repo: tomasonjo/biobert-covidgraph
      dependencies:
        - CORD19
"""
'''
p = Pipeline(description=yaml, docker_client=d, dot_env_path="pipeline.env")
p.add_global_env_var("MY_VAR", "IN_ALL_CONTAINERS")
p.add_global_env_var("MY_JSON", {"my_crazy_dict": "in_all_containers"})
# run all containers once
p.run()

# Optional define custom service schedule (https://schedule.readthedocs.io)
# default is once a day at 00:00
#p.service_schedule = schedule.every(30).seconds.do(p.run_service_containers)

# Step into service mode
#p.start_service_mode()

# now servicecontainer01 will run every 30 seconds



import os
import yaml
import json
from dataclasses import dataclass, field, fields
from typing import List, Dict, Union
import logging
from dotenv import load_dotenv
from expandvars import expandvars

log = logging.getLogger(__name__)


@dataclass
class ContainerRegistryItem:
    name: str
    image_repo: str
    tag: str = "latest"
    info_link: str = None
    desc: str = None
    image_reg_password: str = None
    image_reg_username: str = None
    is_service_container: bool = False
    env_vars: Dict = field(default_factory=dict)
    dependencies: List[str] = field(default_factory=list)
    exlude_in_env: List[str] = field(default_factory=list)
    volumes: Dict = field(default_factory=dict)
    command: Dict = field(default_factory=list)
    sidecars: List["ContainerRegistryItem"] = field(default_factory=list)
    labels: Dict[str, Union[str, bool, float, int]] = field(default_factory=dict)

    # alpha feature properties. not documented yet. still under review:

    exlude_in_env: List = field(default_factory=list)

    @classmethod
    def from_dict(cls, d: dict) -> "ContainerRegistryItem":
        if (
            not "name" in d
            or not "image_repo" in d
            or d["name"] is None
            or d["image_repo"] is None
        ):
            raise ValueError(
                "Incomplete container description. Expected at least 'name' and 'image_repo'. Got:\n\t{}".format(
                    d
                )
            )
        regitem = cls(d["name"], d["image_repo"])
        for data_field in fields(cls):
            if data_field.name in d:
                if data_field.name == "sidecars":
                    sidecars = []
                    for sidecar_dict in d[data_field.name]:
                        print(sidecar_dict)
                        sidecars.append(cls(**sidecar_dict))
                    setattr(regitem, data_field.name, sidecars)
                else:
                    setattr(regitem, data_field.name, d[data_field.name])

        return regitem

    def get_env_vars(self, extra_envs: Dict):
        if extra_envs is None:
            extra_envs = {}
        all_envs = {}
        for key, val in {**extra_envs, **self.env_vars}.items():
            # add all keys redudant for with a CONFIGS_ prefix to make env vars compatible with https://git.connect.dzd-ev.de/dzdtools/pythonmodules/-/tree/master/Configs
            all_envs[key] = val
            all_envs["CONFIGS_{}".format(key)] = val
        return all_envs

    def get_labels(self, extra_labels: Dict):
        if extra_labels is None:
            extra_labels = {}
        all_labels = {}
        for key, val in {**extra_labels, **self.labels}.items():
            all_labels[str(key)] = str(val)
        return all_labels


class ContainerRegistry:
    DATALOADER_BLOCK_LIST: List = None
    DATALOADER_ALLOW_LIST: List = None
    global_env_vars: Dict = None
    global_labels: Dict = None

    name = None

    def __init__(self, data: str, env_file_path: str):
        """[summary]

        Args:
            data (str): File path to yaml or json file OR yaml or json string OR dict
            env_vars (str): File path to a .env file
        """
        self.global_env_vars = {}
        self.parse_dot_env(env_file_path)
        self.items: List[ContainerRegistryItem] = self.parse_data(data)

    def add_global_env(self, key, value):
        log.debug("Add global env '{}' with value '{}'".format(key, value))
        self.global_env_vars[key] = value
        # redudant entry with "CONFIGS_" prefix to support https://git.connect.dzd-ev.de/dzdtools/pythonmodules/-/tree/master/Configs
        if not key.startswith("CONFIGS_"):
            self.global_env_vars["CONFIGS_".format(key)] = value

    def parse_dot_env(self, env_path):
        if not os.path.isfile(env_path):
            log.warning(
                "Can not find a .env file at '{}'. Will skip .env file parsing.".format(
                    os.path.abspath(env_path)
                )
            )
            return
        # load into system envs to have affect while pipeline description parsing
        load_dotenv(dotenv_path=env_path)
        # add these env vars to global env list to be able to pass them into the pipeline children
        with open(env_path) as f:
            for line in f:
                # skip comments, empty lines, broken lines
                if line.startswith("#") or not line.strip() or "=" not in line:
                    continue
                key, value = line.strip().split("=", 1)
                self.add_global_env(key, value)

    def parse_data(self, data) -> List[ContainerRegistryItem]:
        payload = None
        if os.path.isfile(data):
            # we have a json or yaml file
            with open(data, "r") as file:
                payload = "".join(file.readlines())
        elif isinstance(data, str):
            # we have a json or yaml str
            payload = data
        elif isinstance(payload, list):
            d = data
        if payload is not None:
            # Try to parse as yaml. if this failes try to parse a json
            try:
                d = yaml.load(payload, Loader=yaml.FullLoader)
            except Exception as ye:
                try:
                    d = json.load(payload)[list(d.keys())[0]]
                except Exception as je:
                    raise RegistrySourceInvalidEx(
                        "Registry source is not a valid dict, json or yaml file or string.\nYaml parsing error:{}\nJson parsing error:{}".format(
                            ye, je
                        )
                    )
            self.name = list(d.keys())[0]
            d = d[self.name]
            d = self._expand_reg_data_env_vars(d)
            log.debug("Expanded Pipeline:\n{}".format(json.dumps(d, indent=4)))
        registry_item_list = []
        for regitem_dict in d:
            registry_item_list.append(ContainerRegistryItem.from_dict(regitem_dict))
        return registry_item_list

    def _expand_reg_data_env_vars(self, data: Union[Dict, List]) -> Union[Dict, List]:
        # replace all env vars strings (e.g. $MYENVVAR, ${MY_OTHER_ENV_VAR}) by their actual value
        expanded_data = None
        if isinstance(data, list):
            expanded_data = []
            for item in data:
                if isinstance(item, (dict, list)):
                    expanded_data.append(self._expand_reg_data_env_vars(item))
                elif isinstance(item, str):
                    expanded_data.append(expandvars(item))
                else:
                    expanded_data.append(item)
        elif isinstance(data, dict):
            expanded_data = {}
            for key, val in data.items():

                expanded_key = expandvars(key)
                if expanded_key != key:
                    log.debug("EXPAND KEY:'{}' -> '{}'".format(key, expanded_key))
                expanded_data[expanded_key] = val
                if isinstance(val, (dict, list)):
                    expanded_data[expanded_key] = self._expand_reg_data_env_vars(val)
                elif isinstance(val, str):
                    new_val = expandvars(val)
                    if new_val != val:
                        log.debug("EXPAND VALUE:'{}' -> '{}'".format(val, new_val))
                    expanded_data[expanded_key] = new_val
                else:
                    expanded_data[expanded_key] = val
        else:
            raise ValueError("Expected dict or list got {}".format(type(data)))
        return expanded_data

    def get_ordered_containers(self, service_containers_only=False):
        """Returns the container registry sorted by dependencies and filters out reg_items not relevant for the current environment"""
        sorted_reg_items = []
        current_env = os.getenv("ENV", None)
        reg_items_unsorted = self.items

        if self.DATALOADER_BLOCK_LIST and self.DATALOADER_ALLOW_LIST:
            raise ValueError(
                "'DATALOADER_BLOCK_LIST' and 'DATALOADER_ALLOW_LIST' can not be used concurrent. Set one to 'None' or '[]'"
            )
        elif self.DATALOADER_BLOCK_LIST:
            # filter out non listed reg_items
            reg_items_unsorted = [
                reg_item
                for reg_item in reg_items_unsorted
                if reg_item.name not in self.DATALOADER_BLOCK_LIST
            ]
        elif self.DATALOADER_ALLOW_LIST:
            reg_items_unsorted = [
                reg_item
                for reg_item in reg_items_unsorted
                if reg_item.name in self.DATALOADER_ALLOW_LIST
            ]

        def add_data_source(reg_item):
            for index, dep in enumerate(reg_item.dependencies):
                try:
                    add_data_source(
                        next(
                            r_item
                            for r_item in reg_items_unsorted
                            if r_item.name == dep
                        )
                    )
                except StopIteration:
                    log.warning(
                        "Unknown dependency '{}' declared for '{}'. Please check your pipeline defintion. The dependency will be ignored...".format(
                            dep, reg_item.name
                        )
                    )
            if (
                next(
                    (
                        r_item
                        for r_item in sorted_reg_items
                        if r_item.name == reg_item.name
                    ),
                    None,
                )
                is None
                and current_env not in reg_item.exlude_in_env
            ):
                if not service_containers_only or (
                    service_containers_only and reg_item.is_service_container
                ):
                    sorted_reg_items.append(reg_item)

        for r_item in reg_items_unsorted:
            add_data_source(r_item)
        return sorted_reg_items


class RegistrySourceInvalidEx(Exception):
    pass



from setuptools import setup
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="copili",
    description="Run a series of docker/podman containers, in a coordinated manner",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.connect.dzd-ev.de/dzdpythonmodules/copili",
    author="TB",
    author_email="tim.bleimehl@helmholtz-muenchen.de",
    license="MIT",
    packages=["copili"],
    install_requires=["linetimer", "docker", "schedule", "pyyaml", "python-dotenv","expandvars"],
    python_requires=">=3.6",
    zip_safe=False,
    include_package_data=True,
    use_scm_version={
        "root": ".",
        "relative_to": __file__,
        # "local_scheme": "node-and-timestamp"
        "local_scheme": "no-local-version",
        "write_to": "version.py",
    },
    setup_requires=["setuptools_scm"],
)



from .container_manager import ContainerManager
from .registry import ContainerRegistryItem
from .registry import ContainerRegistry
import os
import sys
import logging
import docker
import time
import schedule
import atexit
import signal
from linetimer import CodeTimer
from typing import List, Dict, Callable


log = logging.getLogger(__name__)


if __name__ == "__main__":
    SCRIPT_DIR = os.path.dirname(
        os.path.realpath(os.path.join(os.getcwd(), os.path.expanduser(__file__)))
    )

    sys.path.append(os.path.normpath(SCRIPT_DIR))


class Pipeline:
    registry: ContainerRegistry = None
    global_env_vars: Dict = None
    container_pre_pull_callback: Callable = None
    container_pre_run_callback: Callable = None
    container_post_run_callback: Callable = None
    container_did_run_check_override_callback: Callable = None
    container_dependency_check_override_callback: Callable = None
    container_log_path: str = None
    container_force_image_pull: bool = None
    service_schedule = None
    network_mode = "bridge"

    def __init__(
        self,
        description: str,
        docker_client: docker.DockerClient = None,
        dot_env_path: str = "./.env",
    ):
        """[summary]

        Args:
            description (str): path to yaml or json file or yaml or json string. List of container descriptions
        """
        if docker_client is None:
            self.docker_client = docker.DockerClient(
                base_url="unix://var/run/docker.sock"
            )
        else:
            self.docker_client = docker_client
        self.container_force_image_pull = False
        self.registry: ContainerRegistry = ContainerRegistry(description, dot_env_path)
        self.container_managers: List[ContainerManager] = []
        self.running_container_managers: List[ContainerManager] = []
        self.finished_container_managers: List[ContainerManager] = []
        self.global_env_vars = {}
        self.global_labels = {"copili.dzd-ev.de/container":"true"}
        atexit.register(self._atexit)
        signal.signal(signal.SIGTERM, self._atexit)
        signal.signal(signal.SIGINT, self._atexit)

    def _atexit(self):
        for container_manager in self.running_container_managers:
            container_manager.clean_up()

    def add_global_env_var(self, key, value):
        self.global_env_vars[key] = value

    def add_global_container_label(self,key:str,value:str):
        self.global_labels[key]=value

    def _container_depenencies_did_run_check(self, container_manager: ContainerManager):
        log.info(
            "Check dependencies for {}:".format(container_manager.container_desc.name)
        )
        if callable(self.container_dependency_check_override_callback):
            return self.container_dependency_check_override_callback(
                container_manager,
                self.finished_container_managers,
            )
        else:
            # default dependency check
            successfully_run_reg_item_names = [
                manag.container_desc.name
                for manag in self.finished_container_managers
                if manag.exit_code == 0
            ]

            for reg_item_dep in container_manager.container_desc.dependencies:
                if reg_item_dep not in successfully_run_reg_item_names:
                    return False
        return True

    def _container_did_already_run_check(self, container_manager: ContainerManager):
        # custom caller check
        if callable(self.container_did_run_check_override_callback):
            return self.container_did_run_check_override_callback(
                container_manager, self.finished_container_managers
            )
        else:
            return False

    def _run(self, container_descs: List[ContainerRegistryItem]):
        for container_desc in container_descs:
            skip_or_cancel = False
            container_manager = ContainerManager(
                docker_client=self.docker_client,
                container_desc=container_desc,
                log_path=self.container_log_path,
                force_repull=self.container_force_image_pull,
            )
            self.container_managers.append(container_manager)
            self.running_container_managers.append(container_manager)

            if callable(self.container_pre_pull_callback):
                self.container_pre_pull_callback(container_manager)
            container_manager.pull()

            # skip running container if caller set a skip check function which returns 'True' and container is not a service container
            if (
                not container_desc.is_service_container
                and self._container_did_already_run_check(container_manager)
            ):
                container_manager.exit_code = 0
                container_manager.skipped = True
                log.info(
                    "Skip {} because it did allready run.".format(
                        container_desc.name, ",".join(container_desc.dependencies)
                    )
                )
                skip_or_cancel = True
            else:
                container_manager.skipped = False

            if not self._container_depenencies_did_run_check(container_manager):
                container_manager.exit_code = 1
                container_manager.canceled = True
                log.error(
                    "Can not start {} because of missing dependency (dependecies: {})".format(
                        container_desc.name, ",".join(container_desc.dependencies)
                    )
                )
                skip_or_cancel = True
            else:
                container_manager.canceled = False
            if not skip_or_cancel:
                if callable(self.container_pre_run_callback):
                    self.container_pre_run_callback(container_manager)

                # RUN main container
                container_manager.run(
                    pull=False,
                    extra_envs=self.global_env_vars,
                    network_mode=self.network_mode,
                    labels=self.global_labels
                )
                if callable(self.container_post_run_callback):
                    self.container_post_run_callback(container_manager)
            self.running_container_managers.remove(container_manager)
            self.finished_container_managers.append(container_manager)
            container_manager.log.info("========================================")
            container_manager.log.info(
                "FAILED: {}".format(
                    "False" if container_manager.exit_code == 0 else "True"
                )
            )
            container_manager.log.info(
                "EXITED with status: {}".format(container_manager.exit_code)
            )

            container_manager.log.info("SKIPPED: {}".format(container_manager.skipped))
            container_manager.log.info(
                "CANCELED: {}".format(container_manager.canceled)
            )
            container_manager.log.info(
                "RUNTIME: {} minutes".format(container_manager.run_time / 60)
            )
            container_manager.log.info("========================================")
            container_manager.log.info("")

    def run(self):
        log.info("=====Start '{}'======".format(self.registry.name))
        self._run(self.registry.get_ordered_containers())

    def run_service_containers(self):
        self._run(self.registry.get_ordered_containers(service_containers_only=True))

    def start_service_mode(self):
        log.info("========START SERVICE MODE===========")
        if self.service_schedule is None:
            self.service_schedule = (
                schedule.every().day.at("00:00").do(self.run_service_containers)
            )
            log.info(
                "No service schedule set by caller. Use default schedule:\n\t{}".format(
                    self.service_schedule
                )
            )
        while True:
            schedule.run_pending()
            time.sleep(1)



import os
import sys
import docker
import logging
import time
import json
import threading
from docker.client import DockerClient
from linetimer import CodeTimer
from .registry import ContainerRegistryItem
from typing import Dict, List


class ContainerManager:
    exit_code: int = None
    run_time: float = None
    skipped: bool = None
    canceled: bool = None
    container_desc: ContainerRegistryItem = None
    image: docker.models.images.Image = None
    container: docker.models.containers.Container = None
    sidecars: List["ContainerManager"] = None
    log_handler: logging.Logger = None

    def __init__(
        self,
        docker_client: docker.DockerClient,
        container_desc: ContainerRegistryItem,
        force_repull: bool = True,
        log_to_seperate_file: bool = True,
        log_path: str = None,
    ):
        self.log = logging.getLogger(container_desc.name)
        self.docker_client = docker_client
        self.container_desc = container_desc
        self.force_repull = force_repull
        self.setup_logger(log_to_seperate_file, log_path)
        self.run_time: int = 0
        self.sidecars: List["ContainerManager"] = [
            self.__class__(
                docker_client=docker_client,
                container_desc=sidecar_desc,
                force_repull=force_repull,
                log_to_seperate_file=False,
            )
            for sidecar_desc in container_desc.sidecars
        ]
        self.sidecar_log_threads: List[threading.Thread] = []

    def setup_logger(self, log_to_seperate_file: bool, log_path: str):
        if log_to_seperate_file:
            if log_path is None:
                main_dir = os.path.dirname(
                    os.path.abspath(sys.modules["__main__"].__file__)
                )
                log_path = os.path.join(main_dir, "logs")
            try:
                os.makedirs(log_path)
            except FileExistsError:
                pass

            self.log_handler = logging.FileHandler(
                os.path.join(log_path, "{}.log".format(self.container_desc.name))
            )
            self.log.addHandler(self.log_handler)

    def pull(self):
        self.log.info(
            "Try to pull image '{}' ...".format(self.container_desc.image_repo)
        )
        if self.force_repull:
            self.log.info("... pull is forced; try to removing existing image ...")
            try:
                self.docker_client.images.remove(
                    "{}:{}".format(
                        self.container_desc.image_repo, self.container_desc.tag
                    )
                )
                self.log.info("... image removed ...")
            except docker.errors.ImageNotFound:
                self.log.info("... image not present ...")
                pass
        pull_params = {}
        pull_params["repository"] = self.container_desc.image_repo
        pull_params["tag"] = self.container_desc.tag

        if (
            self.container_desc.image_reg_password is not None
            and self.container_desc.image_reg_username is not None
        ):
            pull_params["auth_config"] = {
                "username": self.container_desc.image_reg_username,
                "password": self.container_desc.image_reg_password,
            }
        self.log.debug("PULL PARAMS:\n{}".format(json.dumps(pull_params, indent=4)))
        self.image = self.docker_client.images.pull(**pull_params)
        self.log.info(
            "...image '{}' pulled.".format(
                "{}:{}".format(self.container_desc.image_repo, self.container_desc.tag)
            )
        )

    def clean_up(self):
        if self.container is not None:
            # clean up sidecars first
            for sidecar in self.sidecars:
                sidecar.clean_up()

            try:
                c = self.container
                self.log.info("Clean up {}".format(self.container_desc.name))
                try:
                    c.kill()
                except docker.errors.APIError:
                    pass
                try:
                    c.remove()
                except docker.errors.APIError:
                    pass
            except docker.errors.NotFound:
                pass
            self.log.info(
                "Stopped and removed container {} - {}:{}".format(
                    self.container_desc.name,
                    self.container_desc.image_repo,
                    self.container_desc.tag,
                )
            )
            if self.log_handler:
                logging.getLogger().removeHandler(self.log_handler)
            # clean up sidecar log threads
            for sidecar_log_thread in self.sidecar_log_threads:
                sidecar_log_thread.join()

    def run(
        self,
        pull=True,
        extra_envs: Dict = None,
        network_mode="bridge",
        background=False,
        labels=None
    ):
        timer = CodeTimer(name=self.container_desc.name, silent=True, unit="s")
        # start sidecars
        links = {}
        for sidecar in self.sidecars:
            self.log.info(f"Start sidecar {sidecar.container_desc.name}")
            sidecar.run(
                pull=pull,
                extra_envs=extra_envs,
                network_mode=network_mode,
                background=True,
                labels={**self.container_desc.get_labels(extra_labels=labels),"copili.dzd-ev.de/sidecar":"true"}
            )
            links[sidecar.container_desc.name] = sidecar.container_desc.name
        self.log_sidecars()

        if extra_envs is None:
            extra_envs = {}
        with timer:
            if pull:
                self.pull()
            self.log.info(
                "Starting container {} from image {}:{}".format(
                    self.container_desc.name,
                    self.container_desc.image_repo,
                    self.container_desc.tag,
                )
            )
            self.container = self.docker_client.containers.run(
                image="{}:{}".format(
                    self.container_desc.image_repo, self.container_desc.tag
                ),
                environment=self.container_desc.get_env_vars(extra_envs),
                detach=True,
                name=self.container_desc.name,
                hostname=self.container_desc.name,
                volumes=self.container_desc.volumes,
                auto_remove=False,  # Autoremove will prevent us from reliable getting the exit code at the end. we need to remove the container manualy
                command=[str(val) for val in self.container_desc.command],
                network_mode=network_mode,
                links=links,
                labels=self.container_desc.get_labels(extra_labels=labels)
            )
            if not background:
                self.recieve_logs()
                try:
                    time.sleep(1)
                    res = self.container.wait()
                except Exception as error:
                    self.log.error(error)
                    # edge case fix; some race condition or whatever. wait() fails
                    self.log.warning("Could not determine exit code. Default to 0")

                    res = {"Error": "None", "StatusCode": 0}
        if not background:
            self.exit_code = res["StatusCode"]
            self.run_time = timer.took

            self.clean_up()

    def recieve_logs(self):
        try:
            for l in self.container.logs(
                stream=True,
                timestamps=True,
                follow=True,
                stderr=True,
                stdout=True,
            ):
                self.log.info(l.decode())
        except (docker.errors.NotFound, docker.errors.APIError) as e:
            # Edge cases; container exited faster then we could attach.
            # this happens for example with debug/demo containers like "hello-world" with virtually no runtime
            # there where some weird docker.errors.APIError ~"409 Client Error: Conflict" cases...could not reproduce. revisit if necessary
            if type(e) == docker.errors.NotFound:
                self.log.info(
                    f"{prefix}No logs for container {self.container_desc.name}"
                )
            else:
                raise e

    def log_sidecars(self):
        for sidecar in self.sidecars:
            log_thread = threading.Thread(
                target=sidecar.recieve_logs,
                args=(),
            )
            log_thread.daemon = True
            log_thread.start()
            self.sidecar_log_threads.append(log_thread)
